import logging
from typing import Iterator, Optional

import opentelemetry.trace
import opik.context_storage
from opik.api_objects import trace, span
from opik.decorator import span_creation_handler, arguments_helpers
from opik.api_objects import opik_client


LOGGER = logging.getLogger(__name__)

NAME_OF_LLM_SPAN_JUST_STARTED_FROM_OPIK_TRACER = "_OPIK_TRACER_STARTED_LLM_SPAN"


class ADKTracerForOpikContextManagement(opentelemetry.trace.NoOpTracer):
    """
    A patched OpenTelemetry tracer for ADK integration.

    This tracer intercepts OpenTelemetry span creation calls from ADK and converts them
    into Opik spans and traces, providing seamless integration between ADK's telemetry
    system and Opik's tracking capabilities.

    Key Features:
    - Converts OpenTelemetry spans to Opik spans and traces
    - Handles trace lifecycle management automatically
    - Skips internal ADK spans that shouldn't have Opik spans created for them
    - Maintains context storage for nested span relationships
    - Provides special logic for handling LLM spans and their lifecycle. In particular:
        * LLM spans are created by OpikTracer.before_model_callback
        * No other Opik spans are created inside LLM spans (ADK creates tool spans inside LLM spans)
    - Supports the flows in which Opik spans are being created, updated or finalized outside of this class.

    The tracer operates as a no-op from OpenTelemetry's perspective (returning INVALID_SPAN)
    while creating corresponding Opik tracking objects in the background.

    Args:
        opik_client: The Opik client instance used for creating spans and traces

    Attributes:
        _ADK_INTERNAL_SPAN_NAME_SKIP_LIST: List of span names that should be skipped
            during tracking.
    """

    _ADK_INTERNAL_SPAN_NAME_SKIP_LIST = [
        "invocation",  # This is the span that is created by ADK when the invocation is started.
    ]

    def __init__(self, opik_client: opik_client.Opik):
        self._opik_client = opik_client

    def start_span(
        self,
        name: str,
        context: Optional[opentelemetry.trace.Context] = None,
        kind: opentelemetry.trace.SpanKind = opentelemetry.trace.SpanKind.INTERNAL,
        attributes: opentelemetry.trace.types.Attributes = None,
        links: opentelemetry.trace._Links = None,
        start_time: Optional[int] = None,
        record_exception: bool = True,
        set_status_on_exception: bool = True,
    ) -> "opentelemetry.trace.Span":
        return opentelemetry.trace.INVALID_SPAN

    @opentelemetry.util._decorator._agnosticcontextmanager
    def start_as_current_span(
        self,
        name: str,
        context: Optional[opentelemetry.trace.Context] = None,
        kind: opentelemetry.trace.SpanKind = opentelemetry.trace.SpanKind.INTERNAL,
        attributes: opentelemetry.trace.types.Attributes = None,
        links: opentelemetry.trace._Links = None,
        start_time: Optional[int] = None,
        record_exception: bool = True,
        set_status_on_exception: bool = True,
        end_on_exit: bool = True,
    ) -> Iterator["opentelemetry.trace.Span"]:
        trace_to_close_in_finally_block = None
        span_to_close_in_finally_block = None

        current_trace_data = opik.context_storage.get_trace_data()
        current_span_data = opik.context_storage.top_span_data()

        if name in self._ADK_INTERNAL_SPAN_NAME_SKIP_LIST:
            yield opentelemetry.trace.INVALID_SPAN
            return

        if (
            current_span_data is not None
            and _is_externally_created_llm_span_ready_for_immediate_finalization(
                current_span_data
            )
        ):
            # LLM span has already been updated from the OpikTracer.after_model_call.
            # ADK will create tool spans inside the llm spans (which we consider wrong practice),
            # so we manually finalize it here to avoid incorrect span nesting.
            opik.context_storage.pop_span_data(ensure_id=current_span_data.id)
            current_span_data.init_end_time()
            self._opik_client.span(**current_span_data.as_parameters)
            current_span_data = opik.context_storage.top_span_data()

        try:
            if current_trace_data is None:
                trace_to_close_in_finally_block = trace.TraceData(name=name)
                opik.context_storage.set_trace_data(trace_to_close_in_finally_block)
            elif (
                current_span_data is not None
                and _is_externally_created_llm_span_that_just_started(current_span_data)
            ):
                # LLM span has just been created and put into context storage from the OpikTracer.before_model_call
                # Not need to create a new one, just remember it to close it in finally block
                span_to_close_in_finally_block = current_span_data
                yield opentelemetry.trace.INVALID_SPAN
                return
            else:
                start_span_arguments = arguments_helpers.StartSpanParameters(
                    name=name,
                    type="general",
                )

                _, span_to_close_in_finally_block = (
                    span_creation_handler.create_span_respecting_context(
                        start_span_arguments=start_span_arguments,
                        distributed_trace_headers=None,
                    )
                )
                opik.context_storage.add_span_data(span_to_close_in_finally_block)

            yield opentelemetry.trace.INVALID_SPAN
        finally:
            if trace_to_close_in_finally_block is not None:
                self._finalize_trace_if_its_not_finalized_yet(
                    trace_to_close_in_finally_block.id
                )
            elif span_to_close_in_finally_block is not None:
                self._finalize_span_if_its_not_finalized_yet(
                    span_to_close_in_finally_block.id
                )
            else:
                LOGGER.warning(
                    "No span or trace to finalize in ADK tracer. This is unexpected."
                )

    def _finalize_trace_if_its_not_finalized_yet(self, trace_id: str) -> None:
        trace_data = opik.context_storage.pop_trace_data(ensure_id=trace_id)
        if trace_data is not None:
            trace_data.init_end_time()
            self._opik_client.trace(**trace_data.as_parameters)

    def _finalize_span_if_its_not_finalized_yet(self, span_id: str) -> None:
        opik.context_storage.trim_span_data_stack_to_certain_span(span_id)

        span_data = opik.context_storage.pop_span_data(ensure_id=span_id)
        if span_data is not None:
            span_data.init_end_time()
            self._opik_client.span(**span_data.as_parameters)


def _is_externally_created_llm_span_ready_for_immediate_finalization(
    span_data: span.SpanData,
) -> bool:
    return (
        span_data.type == "llm"
        and span_data.name != NAME_OF_LLM_SPAN_JUST_STARTED_FROM_OPIK_TRACER
    )


def _is_externally_created_llm_span_that_just_started(
    span_data: span.SpanData,
) -> bool:
    return (
        span_data.type == "llm"
        and span_data.name == NAME_OF_LLM_SPAN_JUST_STARTED_FROM_OPIK_TRACER
    )



from .helpers import (
    try_get_token_usage,
    try_get_streaming_token_usage,
    try_get_ls_metadata,
)

__all__ = [
    "try_get_token_usage",
    "try_get_streaming_token_usage",
    "try_get_ls_metadata",
]



import threading
import logging
import time
from typing import List, Optional

from . import messages, message_queue, queue_consumer
from .. import synchronization
from .batching import batch_manager
from ..file_upload import base_upload_manager
from .. import _logging

LOGGER = logging.getLogger(__name__)


class Streamer:
    def __init__(
        self,
        queue: message_queue.MessageQueue[messages.BaseMessage],
        queue_consumers: List[queue_consumer.QueueConsumer],
        batch_manager: Optional[batch_manager.BatchManager],
        file_upload_manager: base_upload_manager.BaseFileUploadManager,
    ) -> None:
        self._lock = threading.RLock()
        self._message_queue = queue
        self._queue_consumers = queue_consumers
        self._batch_manager = batch_manager
        self._file_upload_manager = file_upload_manager

        self._drain = False

        self._start_queue_consumers()

        if self._batch_manager is not None:
            self._batch_manager.start()

    def put(self, message: messages.BaseMessage) -> None:
        with self._lock:
            if self._drain:
                return

            if (
                self._batch_manager is not None
                and self._batch_manager.message_supports_batching(message)
            ):
                self._batch_manager.process_message(message)
            elif base_upload_manager.message_supports_upload(message):
                self._file_upload_manager.upload(message)
            else:
                if self._message_queue.accept_put_without_discarding() is False:
                    _logging.log_once_at_level(
                        logging.WARNING,
                        "The message queue size limit has been reached. The new message has been added to the queue, and the oldest message has been discarded.",
                        logger=LOGGER,
                    )
                self._message_queue.put(message)

    def close(self, timeout: Optional[int]) -> bool:
        """
        Stops data sending threads
        """
        with self._lock:
            self._drain = True

        if self._batch_manager is not None:
            self._batch_manager.stop()  # stopping causes adding remaining batch messages to the queue

        self.flush(timeout)
        self._close_queue_consumers()

        return self._message_queue.empty()

    def flush(self, timeout: Optional[float], upload_sleep_time: int = 5) -> bool:
        if self._batch_manager is not None:
            self._batch_manager.flush()

        start_time = time.time()

        synchronization.wait_for_done(
            check_function=lambda: self._all_done(),
            timeout=timeout,
            sleep_time=0.1,
        )

        elapsed_time = time.time() - start_time
        if timeout is not None:
            timeout = timeout - elapsed_time
            if timeout < 0.0:
                timeout = 1.0

        # flushing upload manager is blocking operation
        upload_flushed = self._file_upload_manager.flush(
            timeout=timeout, sleep_time=upload_sleep_time
        )

        return upload_flushed and self._all_done()

    def _all_done(self) -> bool:
        return (
            self.workers_idling()
            and self._message_queue.empty()
            and (self._batch_manager is None or self._batch_manager.is_empty())
        )

    def workers_idling(self) -> bool:
        return all([consumer.idling for consumer in self._queue_consumers])

    def queue_size(self) -> int:
        return self._message_queue.size()

    def _start_queue_consumers(self) -> None:
        for consumer in self._queue_consumers:
            consumer.start()

    def _close_queue_consumers(self) -> None:
        for consumer in self._queue_consumers:
            consumer.close()



# This file was auto-generated by Fern from our API Definition.

import typing

import pydantic
import typing_extensions
from ..core.pydantic_utilities import IS_PYDANTIC_V2, UniversalBaseModel
from ..core.serialization import FieldMetadata


class ServiceTogglesConfig(UniversalBaseModel):
    python_evaluator_enabled: typing_extensions.Annotated[bool, FieldMetadata(alias="pythonEvaluatorEnabled")]
    trace_thread_python_evaluator_enabled: typing_extensions.Annotated[
        bool, FieldMetadata(alias="traceThreadPythonEvaluatorEnabled")
    ]
    guardrails_enabled: typing_extensions.Annotated[bool, FieldMetadata(alias="guardrailsEnabled")]

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



import functools

from opik import semantic_version
from opik.api_objects import opik_client
from . import llm_response_wrapper
from . import litellm_wrappers
from . import adk_tracer_for_opik_context_management

import google.adk.models
from google.adk.models import lite_llm
from google.adk import telemetry as adk_telemetry
from google.adk.agents import base_agent


@functools.lru_cache()
def patch_adk(opik_client: opik_client.Opik) -> None:
    # monkey patch LLMResponse to store usage_metadata
    old_function = google.adk.models.LlmResponse.create
    create_wrapper = llm_response_wrapper.LlmResponseCreateWrapper(old_function)
    google.adk.models.LlmResponse.create = create_wrapper

    if hasattr(lite_llm, "LiteLLMClient") and hasattr(
        lite_llm.LiteLLMClient, "acompletion"
    ):
        lite_llm.LiteLLMClient.acompletion = (
            litellm_wrappers.litellm_client_acompletion_decorator(
                lite_llm.LiteLLMClient.acompletion
            )
        )
    if hasattr(lite_llm, "_model_response_to_generate_content_response"):
        lite_llm._model_response_to_generate_content_response = (
            litellm_wrappers.generate_content_response_decorator(
                lite_llm._model_response_to_generate_content_response
            )
        )

    if semantic_version.SemanticVersion.parse(google.adk.__version__) >= "1.3.0":  # type: ignore
        no_op_opik_tracer = (
            adk_tracer_for_opik_context_management.ADKTracerForOpikContextManagement(
                opik_client
            )
        )

        adk_telemetry.tracer.start_as_current_span = (
            no_op_opik_tracer.start_as_current_span
        )
        adk_telemetry.tracer.start_span = no_op_opik_tracer.start_span

        base_agent.tracer.start_as_current_span = (
            no_op_opik_tracer.start_as_current_span
        )
        base_agent.tracer.start_span = no_op_opik_tracer.start_span



import typing
from typing import Dict, Any, Optional

from . import langchain_usage


class LSMetadata(typing.NamedTuple):
    model: str
    provider: str
    model_type: str


def try_get_token_usage(run_dict: Dict[str, Any]) -> langchain_usage.LangChainUsage:
    if (usage := try_get_streaming_token_usage(run_dict)) is not None:
        return usage

    # try generation_info
    usage_metadata = run_dict["outputs"]["generations"][-1][-1]["generation_info"][
        "usage_metadata"
    ]
    return langchain_usage.LangChainUsage.from_original_usage_dict(usage_metadata)


def try_get_streaming_token_usage(
    run_dict: Dict[str, Any],
) -> Optional[langchain_usage.LangChainUsage]:
    if "message" in run_dict["outputs"]["generations"][-1][-1]:
        usage_metadata = run_dict["outputs"]["generations"][-1][-1]["message"][
            "kwargs"
        ]["usage_metadata"]

        return langchain_usage.LangChainUsage.from_original_usage_dict(usage_metadata)

    return None


def try_get_ls_metadata(run_dict: Dict[str, Any]) -> Optional[LSMetadata]:
    if metadata := run_dict["extra"].get("metadata"):
        model = metadata.get("ls_model_name")
        provider = metadata.get("ls_provider")
        model_type = metadata.get("ls_model_type")
        return LSMetadata(model=model, provider=provider, model_type=model_type)

    return None
